﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Hive.Contracts.Interfaces;
using HeuristicLab.Hive.Contracts.ResponseObjects;
using HeuristicLab.Hive.Contracts.BusinessObjects;
using HeuristicLab.Hive.Experiment.Properties;
using System.ServiceModel;
using HeuristicLab.Hive.Contracts;
using HeuristicLab.DataAccess;
using HeuristicLab.Collections;
using HeuristicLab.Common;

namespace HeuristicLab.Hive.Experiment {
  [Item("Hive Client", "Connects to Hive and lists all submitted experiments by the current user.")]
  [Creatable("Hive")]
  public class HiveClient : Item, IProgressReporter {
    private static object locker = new object();
    private bool currentlyUpdating;

    private ILog log;
    public ILog Log {
      get { return log; }
    }

    private HiveExperimentList hiveExperiments;
    public HiveExperimentList HiveExperiments {
      get { return hiveExperiments; }
      set {
        if (hiveExperiments != value) {
          DeRegisterHiveExperimentsEvents();
          hiveExperiments = value;
          RegisterHiveExperimentsEvent();
          OnHiveExperimentsChanged();
        }
      }
    }

    private bool isProgressing;
    public bool IsProgressing {
      get { return isProgressing; }
      set {
        if (isProgressing != value) {
          isProgressing = value;
          OnIsProgressingChanged();
        }
      }
    }

    private IProgress progress;
    public IProgress Progress {
      get { return progress; }
    }

    private void RegisterHiveExperimentsEvent() {
      if (hiveExperiments != null) {
        hiveExperiments.ItemsRemoved += new CollectionItemsChangedEventHandler<IndexedItem<HiveExperiment>>(hiveExperiments_ItemsRemoved);
      }
    }

    private void DeRegisterHiveExperimentsEvents() {
      if (hiveExperiments != null) {
        hiveExperiments.ItemsRemoved -= new CollectionItemsChangedEventHandler<IndexedItem<HiveExperiment>>(hiveExperiments_ItemsRemoved);
      }
    }

    public HiveClient() {
      this.log = new Log();
    }

    public override Common.IDeepCloneable Clone(Common.Cloner cloner) {
      HiveClient clone = (HiveClient)base.Clone(cloner);
      clone.log = (ILog)cloner.Clone(this.log);
      clone.hiveExperiments = (HiveExperimentList)cloner.Clone(this.hiveExperiments);
      return clone;
    }

    public void UpdateExperimentList() {
      this.progress = new Progress("Downloading HiveExperiments...");
      try {
        IsProgressing = true;
        if (this.HiveExperiments == null) {
          this.HiveExperiments = new HiveExperimentList();
        }
        using (Disposable<IClientFacade> service = ServiceLocator.Instance.ClientFacadePool.GetService()) {
          currentlyUpdating = true;
          ResponseObject<HiveExperimentDtoList> response = service.Obj.GetHiveExperiments();
          progress.Status = "Populating HiveExperiment list...";
          RefreshExperimentList(response.Obj);
          currentlyUpdating = false;
        }
      }
      catch (Exception) {
        this.HiveExperiments = null;
        throw;
      }
      finally {
        IsProgressing = false;
      }
    }

    private void RefreshExperimentList(HiveExperimentDtoList hiveExperiments) {
      foreach (HiveExperimentDto hiveExperimentDto in hiveExperiments) {
        HiveExperiment hiveExperiment = GetHiveExperiment(hiveExperimentDto.Id);
        if (hiveExperiment == null) {
          // not yet there, create new
          this.HiveExperiments.Add(new HiveExperiment(hiveExperimentDto));
        } else {
          // update
          hiveExperiment.UpdateFromDto(hiveExperimentDto);
        }
      }
    }

    private HiveExperiment GetHiveExperiment(Guid hiveExperimentId) {
      return this.HiveExperiments.SingleOrDefault(he => he.HiveExperimentId.Equals(hiveExperimentId));
    }

    private void LogMessage(string message) {
      // HeuristicLab.Log is not Thread-Safe, so lock on every call
      lock (locker) {
        log.LogMessage(message);
      }
    }

    public event EventHandler HiveExperimentsChanged;
    private void OnHiveExperimentsChanged() {
      var handler = HiveExperimentsChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    
    void hiveExperiments_ItemsRemoved(object sender, Collections.CollectionItemsChangedEventArgs<Collections.IndexedItem<HiveExperiment>> e) {
      if (!currentlyUpdating) {
        using (Disposable<IClientFacade> service = ServiceLocator.Instance.ClientFacadePool.GetService()) {
          foreach (IndexedItem<HiveExperiment> item in e.Items) {
            if (item.Value.HiveExperimentId != Guid.Empty) {
              service.Obj.DeleteHiveExperiment(item.Value.HiveExperimentId);
            }
          }
        }
      }
    }

    public event EventHandler IsProgressingChanged;
    private void OnIsProgressingChanged() {
      var handler = IsProgressingChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }
  }
}
